import unittest
import sys
sys.path.append('..')

from mcfw.properties import unicode_property, long_property, typed_property
from mcfw.rpc import serialize_complex_value, parse_complex_value

class Test(unittest.TestCase):

    def test_mix_type_property(self):

        class MyLittleTO(object):
            name = unicode_property('name')
            age = long_property('age')

            def __str__(self):
                return u"%s is %s years old" % (self.name, self.age)

            def __eq__(self, other):
                return self.name == other.name and self.age == other.age

        class MyPetTO(object):
            person = typed_property('person', (unicode, MyLittleTO))
            crew = typed_property('crew', (unicode, MyLittleTO), True)

        felix = MyPetTO()
        felix.person = u"nice"
        felix.crew = [u"bart", u"donatello"]

        ceaser = MyPetTO()
        ceaser.person = MyLittleTO()
        ceaser.person.name = u"ceaser"
        ceaser.person.age = 35
        ceaser.crew = [u"bart", MyLittleTO()]
        ceaser.crew[1].name = u"donatello"
        ceaser.crew[1].age = 34

        for person in ceaser.crew:
            print person

        serialized = serialize_complex_value(ceaser, MyPetTO, False)
        print serialized
        ceaser2 = parse_complex_value(MyPetTO, serialized, False)
        self.assertEquals(ceaser.person.name, ceaser2.person.name)
        self.assertEquals(ceaser.person.age, ceaser2.person.age)
        self.assertEquals(ceaser.crew[0], ceaser2.crew[0])
        self.assertEquals(ceaser.crew[1], ceaser2.crew[1])


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

import os


class MissingClass(object):
    def __reduce__(self):
        # See pickle documentation:
        #
        # If a string is returned, the string should be interpreted as the name of a global variable.
        # It should be the objects local name relative to its module; the pickle module searches the module
        # namespace to determine the objects module. This behaviour is typically useful for singletons.
        return 'MISSING'


MISSING = MissingClass()
del MissingClass

SERVER_SOFTWARE = os.environ.get('SERVER_SOFTWARE', 'Development')
DEBUG = SERVER_SOFTWARE.startswith('Development')

AUTHENTICATED = 'authenticated'
NOT_AUTHENTICATED = 'non_authenticated'



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

import base64
import hashlib
import logging
import threading
import time
from collections import defaultdict
from functools import wraps
try:
    from google.appengine.api import memcache as mod_memcache
    from google.appengine.ext import ndb
    __GAE__ = True
except ImportError:  # Allow running outside google app engine
    __GAE__ = False

from mcfw.consts import MISSING
from mcfw.serialization import serializer, s_bool, get_serializer, s_any, deserializer, get_deserializer, ds_bool, \
    ds_any, SerializedObjectOutOfDateException, get_list_serializer, List

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

CACHE_ATTR = u'cache_key'


class CachedModelMixIn(object):
    on_trans_committed = None

    def _trigger_invalidate_cache(self):
        def invalidate_cache():
            self.invalidateCache()
            logging.info('%s: Cache invalidated', self.__class__.__name__)

        if ndb.in_transaction() and self.on_trans_committed:
            self.on_trans_committed(invalidate_cache)
        else:
            invalidate_cache()

    def put(self):
        super(CachedModelMixIn, self).put()
        self._trigger_invalidate_cache()

    def delete(self):
        super(CachedModelMixIn, self).delete()
        self._trigger_invalidate_cache()


class _TLocal(threading.local):
    def __init__(self):
        self.request_cache = dict()


_tlocal = _TLocal()
del _TLocal


def get_from_request_cache(key):
    return _tlocal.request_cache.get(key, MISSING)


def add_to_request_cache(key, success, value):
    _tlocal.request_cache[key] = (success, value)


def flush_request_cache():
    _tlocal.request_cache.clear()


def set_cache_key(wrapped, f):
    def key():
        if hasattr(f, 'meta') and CACHE_ATTR in f.meta:
            return f.meta[CACHE_ATTR]
        else:
            return '%s.%s' % (f.__name__, f.__module__)

    if not hasattr(wrapped, 'meta'):
        wrapped.meta = {CACHE_ATTR: key()}
        return
    if CACHE_ATTR not in wrapped.meta:
        wrapped.meta[CACHE_ATTR] = key()


def ds_key(version, cache_key):
    return '%s-%s' % (version, hashlib.sha256(cache_key).hexdigest())


if __GAE__:
    class DSCache(ndb.Model):
        creation_timestamp = ndb.IntegerProperty()
        description = ndb.StringProperty(indexed=False)
        value = ndb.BlobProperty()

        @property
        def ds_key(self):
            return self.key.id()

        @classmethod
        def create_key(cls, ds_key):
            return ndb.Key(cls, ds_key)


def invalidate_cache(f, *args, **kwargs):
    f.invalidate_cache(*args, **kwargs)


cache_key_locks = defaultdict(lambda: threading.RLock())


def cached(version, lifetime=600, request=True, memcache=True, key=None, datastore=None):
    """
    Caches the result of the decorated function and returns the cached version if it exists.

    Args:
        version (long):  Cache version, needs to bumped every time the arguments of the cached function change
        lifetime (long): Number of seconds the cached entry remains in memcache after it was created.
        request(bool): Whether it needs to be cached in memory for the current request processing.
        memcache (bool): Whether it needs to be cached in memcache.
        key (str): Function to create cache_key
        datastore (str): Content description of cache object in datastore. Leave none to ommit the datastore cache.

    Raises:
        ValueError: if neither request nor memcache are True
        ValueError: If datastore caching is used but lifetime is not set to 0
    """

    if not request and not memcache and not datastore:
        raise ValueError('Either request or memcache or datastore needs to be True')

    if datastore and lifetime != 0:
        raise ValueError('If datastore caching is used, values other than 0 for lifetime are not permitted.')

    def wrap(f):
        base_cache_key = f.meta[CACHE_ATTR]
        if base_cache_key == 'inner_wrapper.google.appengine.api.datastore':
            raise ValueError('Move @db.non_transactional inside the @cached method')
        f_args = f.meta['fargs']
        f_ret = f.meta['return_type']
        f_pure_default_args_dict = f.meta['pure_default_args_dict']

        if isinstance(f_ret, list):
            f_ret = List(f_ret[0])
        if memcache or datastore:
            result_serializer = get_serializer(f_ret)
            result_deserializer = get_deserializer(f_ret)
        key_function = key
        if not key_function:
            def key_(kwargs):
                stream = StringIO()
                stream.write(base_cache_key)
                kwargt = f.meta['kwarg_types']
                for a in sorted(kwargt.keys()):
                    if a in kwargs:
                        effective_value = kwargs[a]
                    else:
                        effective_value = f_pure_default_args_dict[a]
                    if isinstance(kwargt[a], list):
                        get_list_serializer(get_serializer(kwargt[a][0]))(stream, effective_value)
                    else:
                        get_serializer(kwargt[a])(stream, effective_value)
                return stream.getvalue()

            key_function = key_

        @serializer
        def serialize_result(stream, obj):
            s_bool(stream, obj[0])
            if obj[0]:
                result_serializer(stream, obj[1])
            else:
                s_any(stream, obj[1])

        f.serializer = serialize_result

        @deserializer
        def deserialize_result(stream):
            success = ds_bool(stream)
            if success:
                result = result_deserializer(stream)
            else:
                result = ds_any(stream)
            return success, result

        f.deserializer = deserialize_result

        def cache_key(*args, **kwargs):
            kwargs_ = dict(kwargs)
            kwargs_.update(dict(((f_args[0][i], args[i]) for i in xrange(len(args)))))
            return 'v%s.%s' % (version, base64.b64encode(key_function(kwargs_)))

        f.cache_key = cache_key

        def invalidate_cache(*args, **kwargs):
            ck = cache_key(*args, **kwargs)
            with cache_key_locks[ck]:
                if datastore:
                    @ndb.non_transactional
                    def clear_dscache():
                        DSCache.create_key(ds_key(version, ck)).delete()

                    clear_dscache()
                if memcache:
                    attempt = 1
                    while not mod_memcache.delete(ck):  # @UndefinedVariable
                        if attempt >= 3:
                            logging.critical('MEMCACHE FAILURE !!! COULD NOT INVALIDATE CACHE !!!')
                            raise RuntimeError('Could not invalidate memcache!')
                        logging.debug('Memcache failure. Retrying to invalidate cache.')
                        time.sleep(0.25 * attempt)
                        attempt += 1

                if request and ck in _tlocal.request_cache:
                    del _tlocal.request_cache[ck]

        f.invalidate_cache = invalidate_cache

        @wraps(f)
        def wrapped(*args, **kwargs):
            if ndb.in_transaction():
                return f(*args, **kwargs)
            ck = cache_key(*args, **kwargs)
            with cache_key_locks[ck]:
                if request and ck in _tlocal.request_cache:
                    success, result = _tlocal.request_cache[ck]
                    if success:
                        return result
                if memcache:
                    memcache_result = mod_memcache.get(ck)  # @UndefinedVariable
                    if memcache_result:
                        buf = StringIO(memcache_result)
                        try:
                            success, result = deserialize_result(buf)
                            if request:
                                _tlocal.request_cache[ck] = (success, result)
                            if success:
                                return result
                        except SerializedObjectOutOfDateException:
                            pass
                if datastore:
                    @ndb.non_transactional
                    def get_from_dscache():
                        ds_cache = DSCache.create_key(ds_key(version, ck)).get()
                        if ds_cache:
                            buf = StringIO(str(ds_cache.value))
                            try:
                                success, result = deserialize_result(buf)
                                if request:
                                    _tlocal.request_cache[ck] = (success, result)
                                if memcache:
                                    mod_memcache.set(ck, ds_cache.value, time=lifetime)  # @UndefinedVariable
                                if success:
                                    return True, result
                            except SerializedObjectOutOfDateException:
                                pass
                        return False, None

                    cached, result = get_from_dscache()
                    if cached:
                        return result

                cache_value = None
                try:
                    result = f(*args, **kwargs)
                    cache_value = (True, result)
                    return result
                except Exception as e:
                    cache_value = (False, e)
                    raise
                finally:
                    if cache_value and cache_value[0]:
                        if datastore or memcache:
                            buf = StringIO()
                            serialize_result(buf, cache_value)
                            serialized_cache_value = buf.getvalue()
                        if datastore:
                            @ndb.non_transactional
                            def update_dscache():
                                dsm = DSCache(key=DSCache.create_key(ds_key(version, ck)))
                                dsm.description = datastore
                                dsm.creation_timestamp = int(time.time())
                                dsm.value = serialized_cache_value
                                dsm.put()

                            update_dscache()
                        if memcache:
                            mod_memcache.set(ck, serialized_cache_value, time=lifetime)  # @UndefinedVariable
                        if request:
                            _tlocal.request_cache[ck] = cache_value

        return wrapped

    return wrap



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

import httplib


class HttpException(Exception):
    http_code = 0

    def __init__(self, error=None, data=None, **kwargs):
        self.data = data or {}
        if not error and self.http_code in httplib.responses:
            error = httplib.responses[self.http_code]
        self.error = error
        super(HttpException, self).__init__(self, error, **kwargs)


class HttpBadRequestException(HttpException):
    http_code = httplib.BAD_REQUEST

    def __init__(self, *args, **kwargs):
        super(HttpBadRequestException, self).__init__(*args, **kwargs)


class HttpUnAuthorizedException(HttpException):
    http_code = httplib.UNAUTHORIZED

    def __init__(self, *args, **kwargs):
        super(HttpUnAuthorizedException, self).__init__(*args, **kwargs)


class HttpForbiddenException(HttpException):
    http_code = httplib.FORBIDDEN

    def __init__(self, *args, **kwargs):
        super(HttpForbiddenException, self).__init__(*args, **kwargs)


class HttpNotFoundException(HttpException):
    http_code = httplib.NOT_FOUND

    def __init__(self, *args, **kwargs):
        super(HttpNotFoundException, self).__init__(*args, **kwargs)


class HttpConflictException(HttpException):
    http_code = httplib.CONFLICT

    def __init__(self, *args, **kwargs):
        super(HttpConflictException, self).__init__(*args, **kwargs)


class HttpUnprocessableEntityException(HttpException):
    http_code = httplib.UNPROCESSABLE_ENTITY

    def __init__(self, *args, **kwargs):
        super(HttpUnprocessableEntityException, self).__init__(*args, **kwargs)


class HttpInternalServerErrorException(HttpException):
    http_code = httplib.INTERNAL_SERVER_ERROR

    def __init__(self, *args, **kwargs):
        super(HttpInternalServerErrorException, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

import httplib
import inspect
import json
import logging
import threading
import webapp2
from collections import defaultdict
from types import NoneType

from consts import DEBUG, AUTHENTICATED, NOT_AUTHENTICATED
from mcfw.exceptions import HttpException, HttpBadRequestException
from mcfw.properties import simple_types
from mcfw.rpc import run, ErrorResponse, serialize_complex_value, MissingArgumentException, parse_complex_value

DEFAULT_API_VERSION = 'v1.0'

_rest_handlers = defaultdict(dict)
_precall_hooks = []
_postcall_hooks = []


class BadRequestResponse(Exception):
    pass


class InjectedFunctions(object):
    def __init__(self):
        self._get_session_function = None
        self._api_url_template = None

    @property
    def get_current_session(self):
        return self._get_session_function

    @get_current_session.setter
    def get_current_session(self, function):
        self._get_session_function = function

    @property
    def get_api_url_template(self):
        return self._api_url_template

    @get_api_url_template.setter
    def get_api_url_template(self, function):
        self._api_url_template = function


INJECTED_FUNCTIONS = InjectedFunctions()
del InjectedFunctions


def register_precall_hook(callable_):
    _precall_hooks.append(callable_)


def register_postcall_hook(callable_):
    _postcall_hooks.append(callable_)


def rest(uri, method='get', scopes=None, version=DEFAULT_API_VERSION, uri_prefix=None, silent=False,
         silent_result=False):
    if method not in ('get', 'post', 'put', 'delete'):
        ValueError('method')
    if scopes is None:
        scopes = []
    if isinstance(scopes, str):
        scopes = [scopes]

    def wrap(f):
        if not inspect.isfunction(f):
            raise ValueError('%s is not of type function!' % f)

        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        wrapped.__name__ = f.__name__

        api_url = INJECTED_FUNCTIONS.get_api_url_template(version, uri)
        if uri_prefix:
            api_url = "/%s%s" % (uri_prefix, api_url)

        wrapped.meta = {
            'rest': True,
            'uri': api_url,
            'scopes': scopes,
            'method': method,
            "silent": silent,
            "silent_result": silent_result
        }
        if hasattr(f, 'meta'):
            wrapped.meta.update(f.meta)
        return wrapped

    return wrap


class ResponseTracker(threading.local):
    def __init__(self):
        self.current_response = None
        self.current_request = None


_current_reponse_tracker = ResponseTracker()
del ResponseTracker


class GenericRESTRequestHandler(webapp2.RequestHandler):
    @staticmethod
    def getCurrentResponse():
        return _current_reponse_tracker.current_response

    @staticmethod
    def getCurrentRequest():
        return _current_reponse_tracker.current_request

    @staticmethod
    def clearCurrent():
        _current_reponse_tracker.current_response = None

    @staticmethod
    def setCurrent(request, response):
        _current_reponse_tracker.current_request = request
        _current_reponse_tracker.current_response = response

    def ctype(self, type_, value):
        if not isinstance(type_, (list, tuple)):
            if type_ == bool:
                return bool(value) and value.lower() == 'true'
            return type_(value)
        elif isinstance(type_, list):
            return [self.ctype(type_[0], item) for item in value.split(',')]
        elif type_ == (str, unicode):
            return unicode(value)
        elif type_ == (int, long):
            return long(value)
        elif type_ == (int, long, NoneType):
            return None if value is None or value == '' else long(value)
        raise NotImplementedError()

    def get_handler(self, method, route):
        """
        Returns the handler associated with the requested URL
         Returns None if no handler was found for this url or when the method is not implemented/allowed
        Args:
            method (string)
            route (webapp2.Route)
        Returns:
            function(callable)
        """
        if route.template in _rest_handlers:
            if method in _rest_handlers[route.template]:
                return _rest_handlers[route.template][method]
            else:
                self.response.set_status(httplib.METHOD_NOT_ALLOWED, httplib.responses[httplib.METHOD_NOT_ALLOWED])
        self.response.set_status(httplib.NOT_FOUND, httplib.responses[httplib.NOT_FOUND])

    def update_kwargs(self, function, kwargs):
        for name, type_ in function.meta['kwarg_types'].iteritems():
            if name in self.request.GET:
                kwargs[name] = self.ctype(type_, self.request.GET[name])

    def get(self, *args, **kwargs):
        GenericRESTRequestHandler.setCurrent(self.request, self.response)
        f = self.get_handler('get', self.request.route)
        if not f:
            return
        self.update_kwargs(f, kwargs)
        self.write_result(self.run(f, args, kwargs))

    def post(self, *args, **kwargs):
        GenericRESTRequestHandler.setCurrent(self.request, self.response)
        f = self.get_handler('post', self.request.route)
        if not f:
            return
        self.update_kwargs(f, kwargs)
        post_data_type = f.meta['kwarg_types'].get('data')
        if post_data_type:
            is_list = type(post_data_type) is list
            if is_list:
                post_data_type = post_data_type[0]
            post_data = json.loads(self.request.body) if self.request.body else {}
            kwargs['data'] = parse_complex_value(post_data_type, post_data, is_list)
        self.write_result(self.run(f, args, kwargs))

    def put(self, *args, **kwargs):
        GenericRESTRequestHandler.setCurrent(self.request, self.response)
        f = self.get_handler('put', self.request.route)
        if not f:
            return
        self.update_kwargs(f, kwargs)
        post_data_type = f.meta['kwarg_types'].get('data')
        if post_data_type:
            is_list = type(post_data_type) is list
            if is_list:
                post_data_type = post_data_type[0]
            post_data = json.loads(self.request.body) if self.request.body else {}
            kwargs['data'] = parse_complex_value(post_data_type, post_data, is_list)
        self.write_result(self.run(f, args, kwargs))

    def delete(self, *args, **kwargs):
        GenericRESTRequestHandler.setCurrent(self.request, self.response)
        f = self.get_handler('delete', self.request.route)
        if not f:
            return
        self.update_kwargs(f, kwargs)
        self.write_result(self.run(f, args, kwargs))

    def write_result(self, result):
        self.response.headers.update({
            'Content-Type': 'application/json'
        })
        if result is not None:
            if type(result) == ErrorResponse:
                self.response.set_status(result.status_code)
                result = serialize_complex_value(result, ErrorResponse, False)
            if DEBUG:
                self.response.out.write(json.dumps(result, indent=2, sort_keys=True))
            else:
                self.response.out.write(json.dumps(result))
        else:
            self.response.set_status(httplib.NO_CONTENT)

    def run(self, f, args, kwargs):
        """
        Args:
            f (any)
            args (tuple)
            kwargs (dict)
        Returns: callable
        """
        if f.meta['authentication'] == AUTHENTICATED:
            session = INJECTED_FUNCTIONS.get_current_session()
            if not session:
                self.abort(httplib.UNAUTHORIZED)

            if f.meta['scopes']:
                scopes = set()
                simple_kwargs = {}
                for kwarg in kwargs:
                    if type(kwargs[kwarg]) in simple_types:
                        simple_kwargs[kwarg] = kwargs[kwarg]

                for scope in f.meta['scopes']:
                    scopes.add(scope.format(**simple_kwargs))

                if not any(scope in scopes for scope in session.scopes):
                    self.abort(httplib.FORBIDDEN)

        for hook in _precall_hooks:
            hook(f, *args, **kwargs)
        try:
            result = run(f, args, kwargs)
        except HttpException as http_exception:
            return ErrorResponse(http_exception)
        except MissingArgumentException as e:
            logging.debug(e)
            return ErrorResponse(HttpBadRequestException(e.message))
        except Exception as e:
            for hook in _postcall_hooks:
                hook(f, False, kwargs, e)
            raise
        for hook in _postcall_hooks:
            hook(f, True, kwargs, result)
        return result


def rest_functions(module, authentication=AUTHENTICATED):
    if authentication not in (AUTHENTICATED, NOT_AUTHENTICATED):
        raise ValueError(authentication)
    for f in (function for (name, function) in inspect.getmembers(module, lambda x: inspect.isfunction(x))):
        if hasattr(f, 'meta') and f.meta.get('rest', False):
            meta_uri = f.meta['uri']
            meta_method = f.meta['method']
            f.meta['authentication'] = authentication
            for uri in (meta_uri if isinstance(meta_uri, (list, tuple)) else (meta_uri,)):
                _rest_handlers[uri][meta_method] = f
                yield uri, GenericRESTRequestHandler



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

# !/usr/bin/env python

import os

import mcfw

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def path(p):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), p)


requirements = []
tests_requirements = requirements + []

setup(name='mcfw',
      version=mcfw.__version__,
      description='Framework for easy caching and rest api calls in google appengine projects',
      classifiers=[
          'Programming Language :: Python :: 2.7',
          'Environment :: Web Environment',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Software Development :: Libraries',
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
      ],
      keywords=['mcfw', 'appengine', 'rest', 'cache'],
      author='Green IT Globe',
      author_email='apps@greenitglobe.com',
      url='https://github.com/rogerthat-platform/mcfw',
      license='Apache 2.0',
      packages=['mcfw'],
      install_requires=requirements,
      tests_require=tests_requirements,
      )



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

import datetime
import time
import types
from functools import wraps
from struct import Struct

try:
    from google.appengine.api import users
    from google.appengine.ext import ndb
    from google.appengine.ext import db
except ImportError:  # Allow running outside google app engine
    pass

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

_serializers = dict()
_ushortStruct = Struct('<H')
_intStruct = Struct('<i')
_longStruct = Struct('<q')
_longLongStruct = Struct('<q')
_doubleStruct = Struct('<d')


class CustomProperty(object):
    pass


def register(type_, serializer, deserializer):
    _serializers[type_] = (serializer, deserializer)


def serialize(type_, obj):
    stream = StringIO()
    _serializers[type_][0](stream, obj)
    return stream.getvalue()


def deserialize(type_, stream):
    if isinstance(stream, (str, unicode)):
        stream = StringIO(stream)
    if isinstance(stream, db.Blob):
        stream = StringIO(str(stream))
    return _serializers[type_][1](stream)


def get_serializer(type_):
    return _serializers[type_][0]


def get_deserializer(type_):
    return _serializers[type_][1]


def serializer(f):
    @wraps(f)
    def wrapped(stream, obj, *args, **kwargs):
        if obj is None:
            stream.write('0')
        else:
            stream.write('1')
            f(stream, obj, *args, **kwargs)

    return wrapped


def deserializer(f):
    @wraps(f)
    def wrapped(stream, *args, **kwargs):
        if stream.read(1) == '0':
            return None
        else:
            return f(stream, *args, **kwargs)

    return wrapped


@serializer
def s_str(stream, obj):
    stream.write(_intStruct.pack(len(obj)))
    stream.write(obj)


@deserializer
def ds_str(stream):
    (size,) = _intStruct.unpack(stream.read(_intStruct.size))
    return stream.read(size)


register(str, s_str, ds_str)


@serializer
def s_unicode(stream, obj):
    s_str(stream, obj.encode('UTF-8'))


@deserializer
def ds_unicode(stream):
    s = ds_str(stream)
    return None if s is None else s.decode('UTF-8')


register(unicode, s_unicode, ds_unicode)


@serializer
def s_bool(stream, obj):
    stream.write('1' if obj else '0')


@deserializer
def ds_bool(stream):
    return stream.read(1) == '1'


register(bool, s_bool, ds_bool)


@serializer
def s_ushort(stream, obj):
    stream.write(_ushortStruct.pack(obj))


@deserializer
def ds_ushort(stream):
    (value,) = _ushortStruct.unpack(stream.read(_ushortStruct.size))
    return value


@serializer
def s_long(stream, obj):
    stream.write(_longStruct.pack(obj))


@deserializer
def ds_long(stream):
    (value,) = _longStruct.unpack(stream.read(_longStruct.size))
    return value


register(int, s_long, ds_long)


@serializer
def s_long_long(stream, obj):
    stream.write(_longLongStruct.pack(obj))


@deserializer
def ds_long_long(stream):
    (value,) = _longLongStruct.unpack(stream.read(_longLongStruct.size))
    return value


register(long, s_long_long, ds_long_long)


@serializer
def s_float(stream, obj):
    stream.write(_doubleStruct.pack(obj))


@deserializer
def ds_float(stream):
    (value,) = _doubleStruct.unpack(stream.read(_doubleStruct.size))
    return value


register(float, s_float, ds_float)


@serializer
def s_datetime(stream, obj):
    s_long(stream, int(time.mktime(obj.timetuple())))


@deserializer
def ds_datetime(stream):
    return datetime.datetime.fromtimestamp(ds_long(stream))


register(datetime.datetime, s_datetime, ds_datetime)


@serializer
def s_key(stream, key):
    s_str(stream, key.urlsafe())


@deserializer
def ds_key(stream):
    return ndb.Key(urlsafe=ds_str(stream))


if 'ndb' in locals():
    register(ndb.Key, s_key, ds_key)


@serializer
def s_any(stream, obj):
    pickle.dump(obj, stream)


@deserializer
def ds_any(stream):
    return pickle.load(stream)


@serializer
def s_user(stream, obj):
    s_unicode(stream, obj.email())


@deserializer
def ds_user(stream):
    return users.User(ds_unicode(stream))


if 'users' in locals():
    register(users.User, s_user, ds_user)


def _get_model_properties(model):
    props = model._properties
    if not hasattr(model, 'ATTRIBUTES_HASH'):
        prop_keys = sorted(props.keys())
        prop_key_hash = hash(','.join(prop_keys))
        model.ATTRIBUTES_HASH = prop_key_hash
        model.MC_ATTRIBUTES = prop_keys
    else:
        prop_keys = model.MC_ATTRIBUTES
        prop_key_hash = model.ATTRIBUTES_HASH
    return prop_key_hash, prop_keys, props


@serializer
def s_model(stream, obj, clazz=None):
    if clazz is None:
        clazz = obj.__class__
    hash_, keys, properties = _get_model_properties(clazz)
    s_long(stream, hash_)
    s_key(stream, obj.key)
    for key in keys:
        prop = properties[key]
        value = getattr(obj, key)
        prop_repeated = prop._repeated
        if prop.__class__ == ndb.StringProperty:
            if prop_repeated:
                s_long(stream, len(value))
                for s in value:
                    s_unicode(stream, s)
            else:
                s_unicode(stream, value)
        elif prop.__class__ == ndb.IntegerProperty:
            if prop_repeated:
                s_long(stream, len(value))
                for i in value:
                    s_long(stream, i)
            else:
                s_long(stream, value)
        elif prop.__class__ == ndb.DateTimeProperty:
            s_datetime(stream, value)
        elif prop.__class__ == ndb.UserProperty:
            if prop_repeated:
                s_long(stream, len(value))
                for u in value:
                    s_user(stream, u)
            else:
                s_str(stream, value.email() if value else None)
        elif prop.__class__ == ndb.BooleanProperty:
            if prop_repeated:
                s_long(stream, len(value))
                for b in value:
                    s_bool(stream, b)
            else:
                s_bool(stream, value)
        elif prop.__class__ == ndb.TextProperty:
            s_unicode(stream, value)
        elif isinstance(prop, CustomProperty):
            prop.get_serializer()(stream, value)
        elif prop.__class__ == ndb.polymodel._ClassKeyProperty:
            continue
        else:
            raise NotImplementedError('Can not serialize %s instances' % prop.__class__)


@deserializer
def ds_model(stream, cls):
    return model_deserializer(stream, cls)


def model_deserializer(stream, cls):
    hash_, keys, properties = _get_model_properties(cls)
    inst_hash = ds_long(stream)
    if hash_ != inst_hash:
        raise SerializedObjectOutOfDateException()
    kwargs = dict()
    model_key = ds_key(stream)
    for property_name in keys:
        prop = properties[property_name]
        prop_repeated = prop._repeated
        if prop.__class__ == ndb.StringProperty:
            if prop_repeated:
                length = ds_long(stream)
                value = [ds_unicode(stream) for _ in xrange(length)]
            else:
                value = ds_unicode(stream)
        elif prop.__class__ == ndb.IntegerProperty:
            if prop_repeated:
                length = ds_long(stream)
                value = [ds_long(stream) for _ in xrange(length)]
            else:
                value = ds_long(stream)
        elif prop.__class__ == ndb.DateTimeProperty:
            value = ds_datetime(stream)
        elif prop.__class__ == ndb.UserProperty:
            if prop_repeated:
                length = ds_long(stream)
                value = [ds_user(stream) for _ in xrange(length)]
            else:
                value = ds_str(stream)
                if value:
                    value = users.User(value)
        elif prop.__class__ == ndb.BooleanProperty:
            if prop_repeated:
                length = ds_long(stream)
                value = [ds_bool(stream) for _ in xrange(length)]
            else:
                value = ds_bool(stream)
        elif prop.__class__ == ndb.TextProperty:
            value = ds_unicode(stream)
        elif isinstance(prop, CustomProperty):
            value = prop.get_deserializer()(stream)
        elif prop.__class__ == ndb.polymodel._ClassKeyProperty:
            continue
        else:
            raise NotImplementedError('Can not deserialize %s instances' % prop.__class__)
        kwargs[property_name] = value

    return cls(key=model_key, **kwargs)


def get_list_serializer(func):
    @serializer
    def s_list(stream, obj):
        if isinstance(obj, types.GeneratorType):
            obj = list(obj)
        stream.write(_intStruct.pack(len(obj)))
        for o in obj:
            func(stream, o)

    return s_list


def get_list_deserializer(func, needsVersionArg=False):
    if needsVersionArg:
        @deserializer
        def ds_list_version(stream, version):
            (size,) = _intStruct.unpack(stream.read(_intStruct.size))
            return [func(stream, version) for _ in xrange(size)]

        return ds_list_version
    else:
        @deserializer
        def ds_list(stream):
            (size,) = _intStruct.unpack(stream.read(_intStruct.size))
            return [func(stream) for _ in xrange(size)]

        return ds_list


class List(object):
    def __init__(self, type_):
        self.type = type_

    def __hash__(self):
        return hash('List') + hash(self.type)

    def __eq__(self, other):
        return hash(self) == hash(other)


s_str_list = get_list_serializer(s_str)
ds_str_list = get_list_deserializer(ds_str)
register(List(str), s_str_list, ds_str_list)

s_unicode_list = get_list_serializer(s_unicode)
ds_unicode_list = get_list_deserializer(ds_unicode)
register(List(unicode), s_unicode_list, ds_unicode_list)

s_bool_list = get_list_serializer(s_bool)
ds_bool_list = get_list_deserializer(ds_bool)
register(List(bool), s_bool_list, ds_bool_list)

s_long_list = get_list_serializer(s_long)
ds_long_list = get_list_deserializer(ds_long)
register(List(long), s_long_list, ds_long_list)

s_float_list = get_list_serializer(s_float)
ds_float_list = get_list_deserializer(ds_float)
register(List(float), s_float_list, ds_float_list)


class SerializedObjectOutOfDateException(Exception):
    pass



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

import hashlib
import inspect
import pprint
import sys
from types import NoneType

from consts import MISSING

try:
    from google.appengine.ext import ndb

    GAE = True
except ImportError:
    GAE = False  # Allow running outside app engine


def rich_str(obj):
    if GAE:
        if isinstance(obj, ndb.Model):
            d = obj.to_dict()
            d['__class__'] = obj.__class__
            for big_data_attr in ('blob', 'picture', 'avatar'):
                if big_data_attr in d:
                    del d[big_data_attr]
            return pprint.pformat(d)
    if isinstance(obj, (dict, list)):
        return pprint.pformat(obj)
    if isinstance(obj, str):
        return obj
    else:
        return unicode(obj).encode('utf8')


def azzert(condition, error_message=''):
    if not condition:
        raise AssertionError(reduce(lambda remainder, item: remainder + '\n  - ' + item[0] + ': ' + rich_str(item[1]),
                                    sorted(sys._getframe(1).f_locals.iteritems()), str(error_message) + '\n  Locals: '))


simple_types = (int, long, str, unicode, bool, float, NoneType)
__none__ = '__none__'


class object_factory(object):
    def __init__(self, subtype_attr_name, subtype_mapping):
        self.subtype_attr_name = subtype_attr_name
        self.subtype_mapping = subtype_mapping

    def get_subtype(self, instance):
        if isinstance(instance, dict):
            subtype_name = instance.get(self.subtype_attr_name)
        else:
            subtype_name = getattr(instance, self.subtype_attr_name, None)

        if subtype_name is None:
            raise ValueError('%s instance has no or empty attribute \'%s\'' % (type(instance), self.subtype_attr_name))

        subtype = self.subtype_mapping.get(subtype_name, None)
        if subtype is None:
            raise ValueError('\'%s\' not found in %s' % (subtype_name, self.subtype_mapping))

        return subtype


class typed_property(object):
    def __init__(self, name, type_, list_=False, doc=None, subtype_attr_name=None, subtype_mapping=None,
                 default=MISSING, hash_serializer=str):
        self.type = type_
        self.list = list_
        self.attr_name = u'_%s' % name
        self.doc = doc
        self.hash_serializer = hash_serializer
        azzert((subtype_attr_name and subtype_mapping) or (not subtype_mapping and not subtype_attr_name),
               'supply both subtype_attr_name and subtype_mapping')
        azzert(not (list_ and (subtype_attr_name or subtype_mapping)),
               'subtype_mapping is not supported in combination with lists')
        self.subtype_attr_name = subtype_attr_name
        self.subtype_mapping = subtype_mapping
        if default != MISSING:
            if list_:
                azzert(hasattr(default, '__iter__') and len(default) == 0,
                       'Only empty list allowed as default value for list properties')
            if default is None:
                azzert(self.type not in ('bool', 'int', 'long', 'float'),
                       'None not allowed for %s properties' % self.type)

        self.default = default

    def __doc__(self):
        return self.doc

    def __get__(self, instance, owner):
        if not instance:
            return self
        else:
            return getattr(instance, self.attr_name, MISSING)

    def __set__(self, instance, value):
        if value != MISSING:
            if self.list:
                if not isinstance(value, (list, tuple, set)):
                    raise ValueError(
                        'Expected [%s] for \'%s\' and got %s - %s!' % (self.type, self.attr_name, type(value), value))
                for i, x in enumerate(value):
                    if self._is_invalid_type(x, instance):
                        raise ValueError('Not all items are from expected type %s! Encountered item at index %s with '
                                         'type %s.' % (unicode(self.type), i, type(x)))
            else:
                if self._is_invalid_type(value, instance):
                    raise ValueError(
                        'Expected %s for \'%s\' and got %s - %s!' % (self.type, self.attr_name, type(value), value))
        setattr(instance, self.attr_name, value)

    def _is_invalid_type(self, value, instance):
        type_ = self.type.get_subtype(value) if isinstance(self.type, object_factory) else self.type
        if value and not isinstance(value, type_):
            return True

        if self.subtype_attr_name and self.subtype_mapping:
            subtype = self.get_subtype(instance)

            if value and not isinstance(value, subtype):
                raise ValueError('Expected %s and got %s - %s!' % (subtype, type(value), value))

        return False

    def get_subtype(self, instance):
        subtype_name = getattr(instance, self.subtype_attr_name, None)
        if subtype_name is None:
            raise ValueError('%s instance has no or empty attribute \'%s\'' % (type(instance), self.subtype_attr_name))

        subtype = self.subtype_mapping.get(subtype_name, None)
        if subtype is None:
            raise ValueError('\'%s\' not found in %s' % (subtype_name, self.subtype_mapping))

        return subtype


unicode_property = unicode_list_property = None
bool_property = bool_list_property = None
long_property = long_list_property = None
float_property = float_list_property = None

_this_mod = locals()


def _generate_properties():
    for type_, list_, name, hash_serializer in (
            (unicode, True, 'unicode_list_property', lambda x: __none__ if x is None else x.encode('utf8')),
            ((int, long), False, 'long_property', str),
            ((int, long), True, 'long_list_property', str),
            ((float, int), False, 'float_property', str),
            ((float, int), True, 'float_list_property', str),
            (bool, False, 'bool_property', str),
            (bool, True, 'bool_list_property', str)):
        def wrap(t, l):
            def init(self, name, doc=None, default=MISSING, hash_serializer=hash_serializer):
                typed_property.__init__(self, name, t, l, doc, default=default, hash_serializer=hash_serializer)

            return init

        global _this_mod
        _this_mod[name] = type(name, (typed_property,), {'__init__': wrap(type_, list_)})


_generate_properties()
del _generate_properties
del _this_mod


class unicode_property(typed_property):
    def __init__(self, name, doc=None, empty_string_is_null=False, default=MISSING):
        typed_property.__init__(self, name, unicode, False, doc, default=default,
                                hash_serializer=lambda x: __none__ if x is None else x.encode('utf8'))
        self._empty_string_is_null = empty_string_is_null

    def __get__(self, instance, owner):
        value = typed_property.__get__(self, instance, owner)
        return None if self._empty_string_is_null and isinstance(value, unicode) and value == u'' else value


simple_property_types = [unicode_property, unicode_list_property, float_property, float_list_property,
                         bool_property, bool_list_property, long_property, long_list_property]

_members_cache = dict()


def get_members(type_):
    if type_ in _members_cache:
        return _members_cache[type]
    else:
        members = sorted(inspect.getmembers(type_, lambda value: isinstance(value, typed_property)), key=lambda x: x[0])
        simple_members = filter(lambda (name, prop): type(prop) in simple_property_types, members)
        complex_members = filter(lambda x: x not in simple_members, members)
        _members_cache[type] = (complex_members, simple_members)
        return complex_members, simple_members


def get_hash(object_):
    def feed(digester, object_):
        if object_ is None:
            digester.update(__none__)
        cms, sms = get_members(type(object_))
        for name, prop in cms:
            value = getattr(object_, name)
            if prop.list:
                for val in value:
                    feed(digester, val)
            else:
                feed(digester, value)
        for name, prop in sms:
            value = getattr(object_, name)
            if value is None:
                digester.update(__none__)
                continue
            if prop.list:
                for val in value:
                    digester.update(prop.hash_serializer(val))
            else:
                digester.update(prop.hash_serializer(value))

    digester = hashlib.md5()
    feed(digester, object_)
    return digester.hexdigest()



# -*- coding: utf-8 -*-
# Copyright 2017 GIG Technology NV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# @@license_version:1.3@@

import inspect
import itertools
import logging
import time
import types
try:
    from google.appengine.ext import ndb
except ImportError:  # Allow running outside google app engine
    pass
from types import NoneType

from cache import set_cache_key
from mcfw.consts import MISSING
from mcfw.properties import get_members, simple_types, object_factory, long_property, unicode_property, typed_property


class ErrorResponse(object):
    status_code = long_property('1')
    error = unicode_property('2')
    data = typed_property('3', dict)

    def __init__(self, rest_exception):
        """
        Args:
            rest_exception (mcfw.exceptions.HttpException):
        """
        self.status_code = rest_exception.http_code
        self.error = u'%s' % rest_exception.error
        self.data = rest_exception.data


class MissingArgumentException(Exception):
    def __init__(self, name, func=None):
        Exception.__init__(self, '%s is a required argument%s!' % (
            name, (' in function %s' % func.func_name) if func else ''))
        self.name = name


def log_access(call=True, response=True):
    def wrap(f):

        def logged(*args, **kwargs):
            if call:
                arg_str = ''
                for i, arg in enumerate(args):
                    arg_str += '  %s: %s\n' % (i, arg)
                kwarg_str = ''
                for kw, arg in kwargs.iteritems():
                    kwarg_str += '  %s: %s\n' % (kw, arg)
                logging.debug(u'%s.%s\nargs:\n%skwargs:\n%s' % (f.__module__, f.__name__, arg_str, kwarg_str))
            start = time.time()
            try:
                result = f(*args, **kwargs)
                if response:
                    end = time.time()
                    logging.debug(
                        u'%s.%s finished in %s seconds returning %s' % (f.__module__, f.__name__, end - start, result))
                return result
            except:
                if response:
                    end = time.time()
                    logging.exception(u'%s.%s failed in %s seconds' % (f.__module__, f.__name__, end - start))
                raise

        set_cache_key(logged, f)
        logged.__name__ = f.__name__
        logged.__module__ = f.__module__
        if hasattr(f, u'meta'):
            logged.meta.update(f.meta)
        return logged

    return wrap


def arguments(**kwarg_types):
    """ The arguments decorator function describes & validates the parameters of the function."""
    map(_validate_type_spec, kwarg_types.itervalues())

    def wrap(f):
        # validate argspec
        if not inspect.isfunction(f):
            raise ValueError('f is not of type function!')
        f_args = inspect.getargspec(f)
        f_args = inspect.ArgSpec([a for a in f_args[0] if a not in ('self', 'cls')], f_args[1], f_args[2], f_args[3])
        f_arg_count = len(f_args[0])
        f_defaults = f_args[3]
        if not f_defaults:
            f_defaults = []
        f_arg_defaults_count = len(f_defaults)
        f_arg_no_defaults_count = f_arg_count - f_arg_defaults_count
        f_arg_defaults = dict(
            (f_args[0][i], f_defaults[i - f_arg_no_defaults_count] if i >= f_arg_no_defaults_count else MISSING) for i
            in xrange(f_arg_count))
        f_pure_default_args_dict = dict((f_args[0][i], f_defaults[i - f_arg_no_defaults_count]) for i in
                                        xrange(f_arg_no_defaults_count, f_arg_count))
        if not f_arg_count == len(kwarg_types):
            raise ValueError(f.func_name + ' does not contain the expected arguments!')
        unknown_args = filter(lambda arg: arg not in kwarg_types, f_args[0])
        if unknown_args:
            raise ValueError('No type information is supplied for %s!' % ', '.join(unknown_args))

        def typechecked_f(*args, **kwargs):
            if len(args) > len(f_args[0]):
                raise ValueError('%s() takes %s arguments (%s given)' % (f.__name__, len(f_args[0]), len(args)))

            kwargs.update(dict(((f_args[0][i], args[i]) for i in xrange(len(args)))))
            # accept MISSING as magical value or not
            accept_missing = u'accept_missing' in kwargs
            if accept_missing:
                kwargs.pop(u'accept_missing')
            # apply default value if available
            for arg, _ in kwarg_types.iteritems():
                value = kwargs.get(arg, f_arg_defaults[arg])
                if value is MISSING:
                    value = f_arg_defaults.get(arg, MISSING)
                kwargs[arg] = value
            # validate number of arguments
            if not len(kwargs) == len(kwarg_types):
                raise ValueError('kwarg mismatch\nExpected:%s\nGot:%s' % (kwarg_types, kwargs))
            # validate supplied arguments
            unknown_args = filter(lambda arg: arg not in kwarg_types, kwargs)
            if unknown_args:
                raise ValueError('Unknown argument(s) %s supplied!' % ', '.join(unknown_args))
            # validate argument values
            map(lambda arg: _check_type(arg, kwarg_types[arg], kwargs[arg], accept_missing=accept_missing, func=f),
                kwargs)
            return f(**kwargs)

        set_cache_key(typechecked_f, f)
        typechecked_f.__name__ = f.__name__
        typechecked_f.__module__ = f.__module__
        typechecked_f.meta[u'fargs'] = f_args
        typechecked_f.meta[u'kwarg_types'] = kwarg_types
        typechecked_f.meta[u'pure_default_args_dict'] = f_pure_default_args_dict
        if hasattr(f, u'meta'):
            typechecked_f.meta.update(f.meta)

        return typechecked_f

    return wrap


def returns(type_=NoneType):
    """ The returns decorator function describes & validates the result of the function."""
    _validate_type_spec(type_)

    def wrap(f):
        def typechecked_return(*args, **kwargs):
            result = f(*args, **kwargs)
            return _check_type(u'Result', type_, result, func=f)

        set_cache_key(typechecked_return, f)
        typechecked_return.__name__ = f.__name__
        typechecked_return.__module__ = f.__module__
        typechecked_return.meta[u'return_type'] = type_
        if hasattr(f, u'meta'):
            typechecked_return.meta.update(f.meta)
        return typechecked_return

    return wrap


def run(function, args, kwargs):
    kwargs['accept_missing'] = None
    result = function(*args, **kwargs)
    type_, islist = _get_return_type_details(function)
    return serialize_value(result, type_, islist, skip_missing=True)


def parse_parameters(function, parameters):
    kwarg_types = get_parameter_types(function)
    return get_parameters(parameters, kwarg_types)


def parse_complex_value(type_, value, islist):
    if value is None:
        return None
    if isinstance(type_, tuple):
        if islist:
            return [_parse_value(None, type_, val) for val in value]
        else:
            return _parse_value(None, type_, value)
    else:
        parser = _get_complex_parser(type_)
        if islist:
            return map(parser, value)
        else:
            return parser(value)


def check_function_metadata(function):
    if 'kwarg_types' not in function.meta or 'return_type' not in function.meta:
        raise ValueError('Can not execute function. Too little meta information is available!')


def get_parameter_types(function):
    return function.meta['kwarg_types']


def get_parameters(parameters, kwarg_types):
    return {name: parse_parameter(name, type_, parameters[name]) if name in parameters else MISSING
            for name, type_ in kwarg_types.iteritems()}


def get_type_details(type_, value=MISSING):
    if isinstance(type_, tuple):
        # The value can have multiple types.
        if value is not MISSING:
            # We must find the type by comparing the possible types with the real type of <value>
            value_is_list = isinstance(value, list)
            if value_is_list:
                if not value:
                    return unicode, True  # The type doesn't matter, the list is empty
                value = value[0]
            for t in type_:
                is_list = isinstance(t, list)
                if is_list != value_is_list:
                    continue
                if is_list:
                    t = t[0]
                if t in (str, unicode):
                    type_to_check = (str, unicode)
                elif t in (int, long):
                    type_to_check = (int, long)
                else:
                    type_to_check = t
                if isinstance(value, type_to_check):
                    return type(value), is_list
                    # Weird... type not found and @arguments didn't raise... The serialization will probably fail.

    is_list = isinstance(type_, list)
    if is_list:
        type_ = type_[0]
    return type_, is_list


def serialize_complex_value(value, type_, islist, skip_missing=False):
    if type_ == dict:
        return value

    def optimal_serializer(val):
        if isinstance(type_, tuple):
            for type_option in type_:
                if isinstance(val, type_option):
                    def serializer(value, skip_missing):
                        return serialize_value(value, type_option, False, skip_missing)
                    break
            else:
                raise ValueError("Could not map val to a type in %s" % type_)
        elif not isinstance(type_, object_factory) and isinstance(val, type_):
            serializer = _get_complex_serializer(val.__class__)
        else:
            serializer = _get_complex_serializer(type_)
        return serializer(val, skip_missing)

    if value is None:
        return None
    if islist:
        try:
            return map(optimal_serializer, value)
        except:
            logging.warn('value for type %s was %s', type_, value)
            raise
    else:
        return optimal_serializer(value)


def serialize_value(value, type_, islist, skip_missing=False):
    if value is None or type_ in simple_types or (isinstance(type_, tuple) and all(t in simple_types for t in type_)):
        return value
    else:
        return serialize_complex_value(value, type_, islist, skip_missing)


def parse_parameter(name, type_, value):
    raw_type, is_list = get_type_details(type_, value)
    if isinstance(value, list) != is_list:
        raise ValueError('list expected for parameter %s and got %s or vice versa!' % (name, value))
    if isinstance(value, list):
        return map(lambda x: _parse_value(name, raw_type, x), value)
    else:
        return _parse_value(name, raw_type, value)


def _validate_type_spec(type_):
    if isinstance(type_, list) and len(type_) != 1:
        raise ValueError('Illegal type specification!')


DICT_KEY_ITERATOR_TYPE = type({}.iterkeys())


def _check_type(name, type_, value, accept_missing=False, func=None):
    if value is MISSING:
        if accept_missing:
            return value
        else:
            raise MissingArgumentException(name, func)

    checktype = (str, unicode) if type_ in (str, unicode) else type_
    checktype = (int, long) if checktype in (int, long) else checktype
    if value is None and (isinstance(checktype, list) or type_ not in (int, long, float, bool)):
        return value

    if isinstance(type_, tuple):
        # multiple types are allowed. checking if value is one of the them.
        errors = []
        for t in type_:
            try:
                return _check_type(name, t, value, accept_missing, func)
            except (ValueError, TypeError) as e:
                errors.append(e)
                continue
        logging.debug('\n\n'.join(map(str, errors)))
        raise ValueError('%s is not of expected type %s! Its type is %s:\n%s' % (name, str(type_), type(value), value))

    if isinstance(checktype, list) and isinstance(value, list):
        checktype = (str, unicode) if checktype[0] in (str, unicode) else checktype[0]

        for i, x in enumerate(value):
            t = checktype.get_subtype(x) if isinstance(checktype, object_factory) else checktype
            if not isinstance(x, t):
                raise ValueError(
                    '%s: Not all items were of expected type %s. Encountered an item at index %s with type %s: %s.'
                    % (name, str(checktype), i, type(x), x))
    elif isinstance(checktype, list) and isinstance(value, (types.GeneratorType, ndb.Query, itertools.chain,
                                                            DICT_KEY_ITERATOR_TYPE)):
        checktype = (str, unicode) if checktype[0] in (str, unicode) else checktype[0]

        def checkStreaming():
            for o in value:
                if not isinstance(o, checktype):
                    raise ValueError('%s: Not all items were of expected type %s' % (name, str(checktype)))
                yield o

        return checkStreaming()
    elif checktype == type and isinstance(value, list):
        if len(value) != 1:
            raise ValueError('%s: unexpected type count (%s)' % (name, len(value)))

        def check(t, i):
            if not isinstance(t, type):
                raise ValueError(
                    '%s: Not all items were of expected type %s. Encountered an item at index %s with type %s: %s.'
                    % (name, str(checktype), i, type(x), x))

        if isinstance(value[0], tuple):
            for i, t in enumerate(value[0]):
                check(t, i)
        else:
            check(value[0], 0)
    else:
        if isinstance(checktype, object_factory):
            checktype = checktype.get_subtype(value)
        try:
            if not isinstance(value, checktype):
                raise ValueError(
                    '%s is not of expected type %s! Its type is %s:\n%s' % (name, str(checktype), type(value), value))
        except TypeError as e:
            raise TypeError('%s\nvalue: %s\nchecktype: %s' % (e.message, value, checktype))
    return value


_complexParserCache = {}


def _get_complex_parser(type_):
    if type_ == dict:
        return lambda x: x
    if type_ not in _complexParserCache:
        def parse(value):
            t = type_.get_subtype(value) if isinstance(type_, object_factory) else type_
            inst = t()

            complex_members, simple_members = get_members(t)
            map(lambda (name, prop): setattr(inst, name, value[name] if name in value else MISSING), simple_members)
            map(lambda (name, prop):
                setattr(inst, name, parse_complex_value(
                    prop.get_subtype(inst) if (prop.subtype_attr_name and prop.subtype_mapping) else prop.type,
                    value[name], prop.list) if name in value else MISSING), complex_members)
            return inst

        _complexParserCache[type_] = parse
        return parse
    else:
        return _complexParserCache[type_]


_value_types = {int, long, float, bool, NoneType}


def _parse_value(name, type_, value):
    def raize():
        if name:
            raise ValueError('Incorrect type received for parameter \'%s\'. Expected %s and got %s (%s).'
                             % (name, type_, type(value), value))
        else:
            raise ValueError('Could not parse %s as %s' % (value, type_))

    istuple = isinstance(type_, tuple)
    if (istuple and set(type_).issubset(_value_types)) or type_ in _value_types:
        if not isinstance(value, type_):
            raize()
        return value
    elif istuple:
        for tt in type_:
            try:
                return _parse_value(name, tt, value)
            except ValueError:
                pass
        raize()
    elif value is None:
        return None
    elif type_ == unicode:
        if not isinstance(value, (str, unicode)):
            raize()
        return value if isinstance(value, unicode) else unicode(value)
    elif type_ == str:
        if not isinstance(value, (str, unicode)):
            raize()
        return value
    elif not isinstance(value, dict):
        raize()
    return parse_complex_value(type_, value, False)


_complex_serializer_cache = {}


def _get_complex_serializer(type_):
    if type_ not in _complex_serializer_cache:
        def serializer(value, skip_missing):
            t = type_.get_subtype(value) if isinstance(type_, object_factory) else type_
            complex_members, simple_members = get_members(t)

            result = dict([(name, getattr(value, name)) for (name, _) in simple_members if
                           not skip_missing or getattr(value, name) != MISSING])

            def _serialize(name, prop):
                attr = getattr(value, name)
                real_type = prop.get_subtype(value) if (prop.subtype_attr_name and prop.subtype_mapping) else prop.type
                serialized_value = serialize_complex_value(attr, real_type, prop.list, skip_missing)
                return (name, serialized_value)

            result.update(dict([_serialize(name, prop) for (name, prop) in complex_members if
                                not skip_missing or getattr(value, name) != MISSING]))

            return result

        _complex_serializer_cache[type_] = serializer
        return serializer
    else:
        return _complex_serializer_cache[type_]


def _get_return_type_details(function):
    return get_type_details(function.meta['return_type'])
